@@ -2,10 +2,21 @@
 # pylint: disable=too-many-arguments
 # Python Modules
 import json
-from typing import Any, Dict, List, Literal, Optional
+from typing import Any, Dict, List, Literal, NamedTuple, Optional
+from urllib.parse import quote
 
 # 3rd Party Modules
 import httpx
+from more_itertools import batched, flatten
+from tenacity import (
+    retry,
+    retry_any,
+    retry_if_exception_message,
+    retry_if_exception_type,
+    stop_after_attempt,
+    wait_chain,
+    wait_fixed,
+)
 
 # Local Modules
 from cofactr.helpers import parse_entities
@@ -19,7 +30,7 @@ from cofactr.schema import (
     schema_to_product,
     schema_to_supplier,
 )
-from cofactr.schema.types import Completion
+from cofactr.schema.types import Completion, PartInV0
 
 Protocol = Literal["http", "https"]
 
@@ -29,6 +40,21 @@ drop_none_values = lambda d: {k: v for k, v in d.items() if v is not None}
 BATCH_LIMIT = 500
 
 
+class RetrySettings(NamedTuple):
+    """Retry settings for GraphAPI methods.
+    TODO: Consider extending to other 5xx errors if and when encountered.
+    """
+
+    reraise: bool = True
+    retry: retry_any = (
+        retry_if_exception_type(httpx.ConnectTimeout)
+        | retry_if_exception_type(httpx.ReadTimeout)
+        | retry_if_exception_message(match=r"Server error '502 Bad Gateway'")
+    )
+    stop: stop_after_attempt = stop_after_attempt(3)
+    wait: wait_chain = wait_chain(*[wait_fixed(wait=wait) for wait in [1, 3, 5]])
+
+
 def get_products(
     url,
     client_id,
@@ -43,11 +69,12 @@ def get_products(
     schema,
     filtering,
     timeout: Optional[int] = None,
+    owner_id: Optional[str] = None,
 ) -> httpx.Response:
     """Get products."""
 
     res = httpx.get(
-        f"{url}/products",
+        f"{url}/products/",
         headers=drop_none_values(
             {
                 "X-CLIENT-ID": client_id,
@@ -56,6 +83,7 @@ def get_products(
         ),
         params=drop_none_values(
             {
+                "owner_id": owner_id,
                 "q": query,
                 "fields": fields,
                 "before": before,
@@ -87,6 +115,7 @@ def get_orgs(
     schema,
     timeout,
     filtering,
+    owner_id,
 ) -> httpx.Response:
     """Get orgs."""
 
@@ -100,6 +129,7 @@ def get_orgs(
         ),
         params=drop_none_values(
             {
+                "owner_id": owner_id,
                 "q": query,
                 "before": before,
                 "after": after,
@@ -128,6 +158,7 @@ def get_suppliers(
     schema,
     timeout,
     filtering,
+    owner_id,
 ) -> httpx.Response:
     """Get orgs."""
 
@@ -141,6 +172,7 @@ def get_suppliers(
         ),
         params=drop_none_values(
             {
+                "owner_id": owner_id,
                 "q": query,
                 "before": before,
                 "after": after,
@@ -163,6 +195,7 @@ class GraphAPI:  # pylint: disable=too-many-instance-attributes
 
     PROTOCOL: Protocol = "https"
     HOST = "graph.cofactr.com"
+    retry_settings = RetrySettings()
 
     def __init__(
         self,
@@ -193,6 +226,12 @@ class GraphAPI:  # pylint: disable=too-many-instance-attributes
 
         return res.json()
 
+    @retry(
+        reraise=retry_settings.reraise,
+        retry=retry_settings.retry,
+        stop=retry_settings.stop,
+        wait=retry_settings.wait,
+    )
     def get_products(
         self,
         query: Optional[str] = None,
@@ -205,6 +244,7 @@ class GraphAPI:  # pylint: disable=too-many-instance-attributes
         schema: Optional[ProductSchemaName] = None,
         filtering: Optional[List[Dict]] = None,
         timeout: Optional[int] = None,
+        owner_id: Optional[str] = None,
     ):
         """Get products.
 
@@ -223,6 +263,7 @@ class GraphAPI:  # pylint: disable=too-many-instance-attributes
             filtering: Filter products.
                 Example: `[{"field":"id","operator":"IN","value":["CCCQSA3G9SMR","CCV1F7A8UIYH"]}]`.
             timeout: Time to wait (in seconds) for the server to issue a response.
+            owner_id: Specifies which private data to access.
         """
 
         if not schema:
@@ -242,18 +283,27 @@ class GraphAPI:  # pylint: disable=too-many-instance-attributes
             schema=schema.value,
             filtering=filtering,
             timeout=timeout,
+            owner_id=owner_id,
         )
 
-        extracted_producs = res.json()
+        extracted_products = res.json()
 
-        Product = schema_to_product[schema]  # pylint: disable=invalid-name
+        # Handle schemas that have parsers.
+        Product = schema_to_product.get(schema)  # pylint: disable=invalid-name
 
-        extracted_producs["data"] = [
-            Product(**data) for data in extracted_producs["data"]
-        ]
+        if Product:
+            extracted_products["data"] = [
+                Product(**data) for data in extracted_products["data"]
+            ]
 
-        return extracted_producs
+        return extracted_products
 
+    @retry(
+        reraise=retry_settings.reraise,
+        retry=retry_settings.retry,
+        stop=retry_settings.stop,
+        wait=retry_settings.wait,
+    )
     def get_products_by_searches(
         self,
         queries: List[str],
@@ -261,6 +311,7 @@ class GraphAPI:  # pylint: disable=too-many-instance-attributes
         force_refresh: bool = False,
         schema: Optional[ProductSchemaName] = None,
         timeout: Optional[int] = None,
+        owner_id: Optional[str] = None,
     ):
         """Search for products associated with each query.
 
@@ -271,6 +322,7 @@ class GraphAPI:  # pylint: disable=too-many-instance-attributes
                 `external`.
             schema: Response schema.
             timeout: Time to wait (in seconds) for the server to issue a response.
+            owner_id: Specifies which private data to access.
 
         Returns:
             A dictionary mapping each MPN to a list of matching products.
@@ -282,25 +334,30 @@ class GraphAPI:  # pylint: disable=too-many-instance-attributes
         if not schema:
             schema = self.default_product_schema
 
+        client_id_param = f"&client_id={quote(owner_id)}" if owner_id else ""
+
         res = httpx.post(
-            f"{self.url}/products",
+            f"{self.url}/batch/products/",
             headers=drop_none_values(
                 {
                     "X-CLIENT-ID": self.client_id,
                     "X-API-KEY": self.api_key,
                 }
             ),
-            json=[
-                {
-                    "method": "GET",
-                    "relative_url": (
-                        f"?q={query}&schema={schema.value}&external={bool(external)}"
-                        f"&force_refresh={force_refresh}"
-                    ),
-                }
-                for query in queries
-            ],
+            json={
+                "batch": [
+                    {
+                        "method": "GET",
+                        "relative_url": (
+                            f"?q={quote(query)}&schema={schema.value}&external={bool(external)}"
+                            f"{client_id_param}&force_refresh={force_refresh}"
+                        ),
+                    }
+                    for query in queries
+                ]
+            },
             timeout=timeout,
+            follow_redirects=True,
         )
 
         res.raise_for_status()
@@ -323,6 +380,12 @@ class GraphAPI:  # pylint: disable=too-many-instance-attributes
 
         return query_to_products
 
+    @retry(
+        reraise=retry_settings.reraise,
+        retry=retry_settings.retry,
+        stop=retry_settings.stop,
+        wait=retry_settings.wait,
+    )
     def get_products_by_ids(
         self,
         ids: List[str],
@@ -330,12 +393,11 @@ class GraphAPI:  # pylint: disable=too-many-instance-attributes
         force_refresh: bool = False,
         schema: Optional[ProductSchemaName] = None,
         timeout: Optional[int] = None,
+        owner_id: Optional[str] = None,
     ):
-        """Get a batch of products.
+        """Get a batch of products by IDs.
 
-        Note:
-            A maximum of 500 IDs can be provided. Any more than that, and the server will return
-            a 422 error. Consider breaking the request into batches.
+        Note: Multiple requests are made if more than 250 IDs are provided.
 
         Args:
             ids: Cofactr product IDs to match on.
@@ -344,36 +406,90 @@ class GraphAPI:  # pylint: disable=too-many-instance-attributes
                 `external`.
             schema: Response schema.
             timeout: Time to wait (in seconds) for the server to issue a response.
+            owner_id: Specifies which private data to access.
         """
 
-        num_requested = len(ids)
-
-        if num_requested > BATCH_LIMIT:
-            raise ValueError(
-                "Too many products requested in one call: Requested"
-                f" {num_requested}, but the limit is {BATCH_LIMIT}."
-            )
+        batch_size = 250
 
         if not schema:
             schema = self.default_product_schema
 
-        extracted_products = self.get_products(
-            external=external,
-            force_refresh=force_refresh,
-            schema=schema,
-            filtering=[{"field": "id", "operator": "IN", "value": ids}],
-            limit=BATCH_LIMIT,
-            timeout=timeout,
+        batched_products = [
+            self.get_products(
+                external=external,
+                force_refresh=force_refresh,
+                schema=schema,
+                filtering=[{"field": "id", "operator": "IN", "value": batched_ids}],
+                limit=batch_size,
+                timeout=timeout,
+                owner_id=owner_id,
+            )
+            for batched_ids in batched(ids, n=batch_size)
+        ]
+
+        products_data = list(
+            flatten([products["data"] for products in batched_products])
         )
 
         id_to_product = parse_entities(
             ids=ids,
-            entities=extracted_products["data"],
+            entities=products_data,
             entity_dataclass=schema_to_product[schema],
         )
 
         return id_to_product
 
+    @retry(
+        reraise=retry_settings.reraise,
+        retry=retry_settings.retry,
+        stop=retry_settings.stop,
+        wait=retry_settings.wait,
+    )
+    def get_canonical_product_ids(
+        self,
+        ids: List[str],
+        timeout: Optional[int] = None,
+        owner_id: Optional[str] = None,
+    ):
+        """Get the canonical product ID for each of the given IDs, which may or may not be
+        deprecated.
+        """
+
+        batch_size = 500
+
+        batched_products = [
+            self.get_products(
+                fields="id,deprecated_ids",
+                external=False,
+                force_refresh=False,
+                schema=ProductSchemaName.INTERNAL,
+                filtering=[{"field": "id", "operator": "IN", "value": batched_ids}],
+                limit=batch_size,
+                timeout=timeout,
+                owner_id=owner_id,
+            )
+            for batched_ids in batched(ids, n=batch_size)
+        ]
+
+        id_to_canonical_id = {}
+
+        for res in batched_products:
+            products = res.get("data", [])
+
+            for product in products:
+                canonical_id = product["id"]
+
+                for key in [canonical_id, *(product.get("deprecated_ids") or [])]:
+                    id_to_canonical_id[key] = canonical_id
+
+        return {id_: id_to_canonical_id.get(id_) for id_ in ids}
+
+    @retry(
+        reraise=retry_settings.reraise,
+        retry=retry_settings.retry,
+        stop=retry_settings.stop,
+        wait=retry_settings.wait,
+    )
     def get_orgs(
         self,
         query: Optional[str] = None,
@@ -383,6 +499,7 @@ class GraphAPI:  # pylint: disable=too-many-instance-attributes
         schema: Optional[OrgSchemaName] = None,
         timeout: Optional[int] = None,
         filtering: Optional[List[Dict]] = None,
+        owner_id: Optional[str] = None,
     ):
         """Get organizations.
 
@@ -395,6 +512,7 @@ class GraphAPI:  # pylint: disable=too-many-instance-attributes
             timeout: Time to wait (in seconds) for the server to issue a response.
             filtering: Filter orgs.
                 Example: `[{"field":"id","operator":"IN","value":["622fb450e4c292d8287b0af5"]}]`.
+            owner_id: Specifies which private data to access.
         """
 
         if not schema:
@@ -411,6 +529,7 @@ class GraphAPI:  # pylint: disable=too-many-instance-attributes
             schema=schema.value,
             timeout=timeout,
             filtering=filtering,
+            owner_id=owner_id,
         )
 
         res_json = res.json()
@@ -421,6 +540,12 @@ class GraphAPI:  # pylint: disable=too-many-instance-attributes
 
         return res_json
 
+    @retry(
+        reraise=retry_settings.reraise,
+        retry=retry_settings.retry,
+        stop=retry_settings.stop,
+        wait=retry_settings.wait,
+    )
     def get_suppliers(
         self,
         query: Optional[str] = None,
@@ -430,6 +555,7 @@ class GraphAPI:  # pylint: disable=too-many-instance-attributes
         schema: Optional[SupplierSchemaName] = None,
         timeout: Optional[int] = None,
         filtering: Optional[List[Dict]] = None,
+        owner_id: Optional[str] = None,
     ):
         """Get suppliers.
 
@@ -442,6 +568,7 @@ class GraphAPI:  # pylint: disable=too-many-instance-attributes
             timeout: Time to wait (in seconds) for the server to issue a response.
             filtering: Filter suppliers.
                 Example: `[{"field":"id","operator":"IN","value":["622fb450e4c292d8287b0af5"]}]`.
+            owner_id: Specifies which private data to access.
         """
 
         if not schema:
@@ -458,6 +585,7 @@ class GraphAPI:  # pylint: disable=too-many-instance-attributes
             schema=schema.value,
             timeout=timeout,
             filtering=filtering,
+            owner_id=owner_id,
         )
 
         res_json = res.json()
@@ -468,56 +596,71 @@ class GraphAPI:  # pylint: disable=too-many-instance-attributes
 
         return res_json
 
+    @retry(
+        reraise=retry_settings.reraise,
+        retry=retry_settings.retry,
+        stop=retry_settings.stop,
+        wait=retry_settings.wait,
+    )
     def get_suppliers_by_ids(
         self,
         ids: List[str],
         schema: Optional[SupplierSchemaName] = None,
         timeout: Optional[int] = None,
+        owner_id: Optional[str] = None,
     ):
-        """Get a batch of suppliers.
+        """Get a batch of suppliers by IDs.
 
-        Note:
-            A maximum of 500 IDs can be provided. Any more than that, and the server will return
-            a 422 error. Consider breaking the request into batches.
+        Note: Multiple requests are made if more than 250 IDs are provided.
 
         Args:
             ids: Cofactr org IDs to match on.
             schema: Response schema.
             timeout: Time to wait (in seconds) for the server to issue a response.
+            owner_id: Specifies which private data to access.
         """
 
-        num_requested = len(ids)
-
-        if num_requested > BATCH_LIMIT:
-            raise ValueError(
-                "Too many suppliers requested in one call: Requested"
-                f" {num_requested}, but the limit is {BATCH_LIMIT}."
-            )
+        batch_size = 250
 
         if not schema:
             schema = self.default_supplier_schema
 
-        extracted_suppliers = self.get_suppliers(
-            schema=schema,
-            filtering=[{"field": "id", "operator": "IN", "value": ids}],
-            limit=BATCH_LIMIT,
-            timeout=timeout,
+        batched_suppliers = [
+            self.get_suppliers(
+                schema=schema,
+                filtering=[{"field": "id", "operator": "IN", "value": batched_ids}],
+                limit=batch_size,
+                timeout=timeout,
+                owner_id=owner_id,
+            )
+            for batched_ids in batched(ids, n=batch_size)
+        ]
+
+        suppliers_data = list(
+            flatten([suppliers["data"] for suppliers in batched_suppliers])
         )
 
         id_to_supplier = parse_entities(
             ids=ids,
-            entities=extracted_suppliers["data"],
+            entities=suppliers_data,
             entity_dataclass=schema_to_supplier[schema],
         )
 
         return id_to_supplier
 
+    @retry(
+        reraise=retry_settings.reraise,
+        retry=retry_settings.retry,
+        stop=retry_settings.stop,
+        wait=retry_settings.wait,
+    )
     def autocomplete_orgs(
         self,
         query: Optional[str] = None,
         limit: Optional[int] = None,
         types: Optional[str] = None,
         timeout: Optional[int] = None,
+        owner_id: Optional[str] = None,
     ) -> Dict[Literal["data"], Completion]:
         """Autocomplete organizations.
 
@@ -532,10 +675,11 @@ class GraphAPI:  # pylint: disable=too-many-instance-attributes
                 Example: "supplier|manufacturer" filters to orgs that are a
                     supplier or a manufacturer.
             timeout: Time to wait (in seconds) for the server to issue a response.
+            owner_id: Specifies which private data to access.
         """
 
         res = httpx.get(
-            f"{self.url}/orgs/autocomplete",
+            f"{self.url}/orgs/autocompletions/",
             headers=drop_none_values(
                 {
                     "X-CLIENT-ID": self.client_id,
@@ -544,6 +688,7 @@ class GraphAPI:  # pylint: disable=too-many-instance-attributes
             ),
             params=drop_none_values(
                 {
+                    "owner_id": owner_id,
                     "q": query,
                     "limit": limit,
                     "types": types,
@@ -557,6 +702,64 @@ class GraphAPI:  # pylint: disable=too-many-instance-attributes
 
         return res.json()
 
+    @retry(
+        reraise=retry_settings.reraise,
+        retry=retry_settings.retry,
+        stop=retry_settings.stop,
+        wait=retry_settings.wait,
+    )
+    def autocomplete_classifications(
+        self,
+        query: Optional[str] = None,
+        limit: Optional[int] = None,
+        types: Optional[str] = None,
+        timeout: Optional[int] = None,
+        owner_id: Optional[str] = None,
+    ) -> Dict[Literal["data"], Completion]:
+        """Autocomplete classifications.
+
+        Args:
+            query: Search query.
+            before: Upper page boundry, expressed as a product ID.
+            after: Lower page boundry, expressed as a product ID.
+            limit: Restrict the results of the query to a particular number of
+                documents.
+            types: Filter for types of organizations.
+                Example: "part_classification" filters to part classification classes.
+            timeout: Time to wait (in seconds) for the server to issue a response.
+            owner_id: Specifies which private data to access.
+        """
+
+        res = httpx.get(
+            f"{self.url}/classes/autocompletions/",
+            headers=drop_none_values(
+                {
+                    "X-CLIENT-ID": self.client_id,
+                    "X-API-KEY": self.api_key,
+                }
+            ),
+            params=drop_none_values(
+                {
+                    "owner_id": owner_id,
+                    "q": query,
+                    "limit": limit,
+                    "types": types,
+                }
+            ),
+            timeout=timeout,
+            follow_redirects=True,
+        )
+
+        res.raise_for_status()
+
+        return res.json()
+
+    @retry(
+        reraise=retry_settings.reraise,
+        retry=retry_settings.retry,
+        stop=retry_settings.stop,
+        wait=retry_settings.wait,
+    )
     def get_product(
         self,
         id: str,
@@ -565,6 +768,7 @@ class GraphAPI:  # pylint: disable=too-many-instance-attributes
         force_refresh: bool = False,
         schema: Optional[ProductSchemaName] = None,
         timeout: Optional[int] = None,
+        owner_id: Optional[str] = None,
     ):
         """Get product.
 
@@ -578,6 +782,7 @@ class GraphAPI:  # pylint: disable=too-many-instance-attributes
                 `external`.
             schema: Response schema.
             timeout: Time to wait (in seconds) for the server to issue a response.
+            owner_id: Specifies which private data to access.
         """
 
         if not schema:
@@ -593,6 +798,7 @@ class GraphAPI:  # pylint: disable=too-many-instance-attributes
             ),
             params=drop_none_values(
                 {
+                    "owner_id": owner_id,
                     "fields": fields,
                     "external": external,
                     "force_refresh": force_refresh,
@@ -615,6 +821,64 @@ class GraphAPI:  # pylint: disable=too-many-instance-attributes
 
         return res_json
 
+    @retry(
+        reraise=retry_settings.reraise,
+        retry=retry_settings.retry,
+        stop=retry_settings.stop,
+        wait=retry_settings.wait,
+    )
+    def create_product(
+        self,
+        data: PartInV0,
+        schema: Optional[ProductSchemaName] = None,
+        timeout: Optional[int] = None,
+    ):
+        """Create product.
+
+        Args:
+            data: Data defining the product to create.
+            schema: Response schema.
+            timeout: Time to wait (in seconds) for the server to issue a response.
+
+        Returns:
+            The newly created product, represented in the given schema.
+        """
+
+        if not schema:
+            schema = self.default_product_schema
+
+        res = httpx.post(
+            f"{self.url}/products/",
+            json=data,
+            headers=drop_none_values(
+                {
+                    "X-CLIENT-ID": self.client_id,
+                    "X-API-KEY": self.api_key,
+                }
+            ),
+            params=drop_none_values({"schema": schema.value}),
+            timeout=timeout,
+            follow_redirects=True,
+        )
+
+        res.raise_for_status()
+
+        res_json = res.json()
+
+        Product = schema_to_product[schema]  # pylint: disable=invalid-name
+
+        res_json["data"] = (
+            Product(**res_json["data"]) if (res_json and res_json.get("data")) else None
+        )
+
+        return res_json
+
+    @retry(
+        reraise=retry_settings.reraise,
+        retry=retry_settings.retry,
+        stop=retry_settings.stop,
+        wait=retry_settings.wait,
+    )
     def get_offers(
         self,
         product_id: str,
@@ -623,6 +887,7 @@ class GraphAPI:  # pylint: disable=too-many-instance-attributes
         force_refresh: bool = False,
         schema: Optional[OfferSchemaName] = None,
         timeout: Optional[int] = None,
+        owner_id: Optional[str] = None,
     ):
         """Get product.
 
@@ -634,6 +899,7 @@ class GraphAPI:  # pylint: disable=too-many-instance-attributes
                 `external`.
             schema: Response schema.
             timeout: Time to wait (in seconds) for the server to issue a response.
+            owner_id: Specifies which private data to access.
         """
 
         if not schema:
@@ -649,6 +915,7 @@ class GraphAPI:  # pylint: disable=too-many-instance-attributes
             ),
             params=drop_none_values(
                 {
+                    "owner_id": owner_id,
                     "fields": fields,
                     "external": external,
                     "force_refresh": force_refresh,
@@ -669,11 +936,18 @@ class GraphAPI:  # pylint: disable=too-many-instance-attributes
 
         return res_json
 
+    @retry(
+        reraise=retry_settings.reraise,
+        retry=retry_settings.retry,
+        stop=retry_settings.stop,
+        wait=retry_settings.wait,
+    )
     def get_org(
         self,
         id: str,
         schema: Optional[OrgSchemaName] = None,
         timeout: Optional[int] = None,
+        owner_id: Optional[str] = None,
     ):
         """Get organization."""
 
@@ -688,7 +962,7 @@ class GraphAPI:  # pylint: disable=too-many-instance-attributes
                     "X-API-KEY": self.api_key,
                 }
             ),
-            params=drop_none_values({"schema": schema.value}),
+            params=drop_none_values({"owner_id": owner_id, "schema": schema.value}),
             timeout=timeout,
             follow_redirects=True,
         )
@@ -705,11 +979,18 @@ class GraphAPI:  # pylint: disable=too-many-instance-attributes
 
         return res_json
 
+    @retry(
+        reraise=retry_settings.reraise,
+        retry=retry_settings.retry,
+        stop=retry_settings.stop,
+        wait=retry_settings.wait,
+    )
     def get_supplier(
         self,
         id: str,
         schema: Optional[SupplierSchemaName] = None,
         timeout: Optional[int] = None,
+        owner_id: Optional[str] = None,
     ):
         """Get supplier."""
 
@@ -724,7 +1005,7 @@ class GraphAPI:  # pylint: disable=too-many-instance-attributes
                     "X-API-KEY": self.api_key,
                 }
             ),
-            params=drop_none_values({"schema": schema.value}),
+            params=drop_none_values({"owner_id": owner_id, "schema": schema.value}),
             timeout=timeout,
             follow_redirects=True,
         )
