@@ -1,4 +1,5 @@
 from cofactr.graph import GraphAPI
+from cofactr.schema.types import PartInV0
 
 graph = GraphAPI(protocol="http", host="127.0.0.1:8050")
 
@@ -6,15 +7,14 @@ graph = GraphAPI(protocol="http", host="127.0.0.1:8050")
 #     data=PartInV0(
 #         owner_id="joseph",
 #         mpn="test",
-#         classification=ClassificationInV0(custom_label="Doodle class 2"),
 #     )
 # )
 
-res = graph.autocomplete_classifications(query="Solar Cell")
+# import pdb
 
-import pdb
+# pdb.set_trace()
 
-pdb.set_trace()
+res = graph.get_offers(product_id="XX9YD048JZIV", owner_id="joseph")
 
 # res = graph.create_product(
 #     data={
